# migration script from seda 2.1 to seda 2.2

import logging

logger = logging.getLogger("seda.migration")
logger.setLevel(logging.INFO)

logger.info("\n\nUpdate schema for seda 2.2 \n\n")

logger.info("\n\nRename entities: \n")

RENAMED_ETYPES = (
    ("SEDAAgentAbstract", "SEDAAgent"),
    ("SEDAAltAgentAbstractCorpname", "SEDAAltAgentCorpname"),
    ("SEDASeqAltAgentAbstractCorpnameFirstName", "SEDASeqAltAgentCorpnameFirstName"),
)

for old_etype, new_etype in RENAMED_ETYPES:
    try:
        eschema = cnx.repo.schema.entity_schema_for(old_etype)  # noqa
    except KeyError:
        # check the new entity exists
        cnx.repo.schema.entity_schema_for(new_etype)  # noqa
        logger.warning(
            f"Entity '{old_etype}' doesn't exist and has been replaced by {new_etype}"
        )
        continue
    count = rql(f"Any COUNT(X) WHERE X is {old_etype}")[0][0]  # noqa
    if count > 0:
        logger.error(f"Found {count} {old_etype}")
        raise
    logger.info(f"Found {count} {old_etype}")
    drop_entity_type(old_etype)  # noqa
    add_entity_type(new_etype)  # noqa

cnx.commit()  # noqa

NEW_ETYPES = (
    "SEDAHoldRuleCodeListVersion",
    "SEDADataObjectProfile",
    "SEDAHoldRule",
    "SEDASeqHoldRuleRule",
    "SEDAAltHoldRulePreventInheritance",
    "SEDAHoldEndDate",
    "SEDAHoldOwner",
    "SEDAHoldReassessingDate",
    "SEDAHoldReason",
    "SEDAPreventRearrangement",
    "SEDADateLitteral",
    "SEDAOriginatingSystemIdReplyTo",
    "SEDATextContent",
    "SEDALinkingAgentIdentifier",
    "SEDALinkingAgentIdentifierType",
    "SEDALinkingAgentIdentifierValue",
    "SEDALinkingAgentRole",
)

logger.info("\n\nAdd new entities: \n")

for new_etype in NEW_ETYPES:
    add_entity_type(new_etype)  # noqa
    logger.info(f"Add new entity type '{new_etype}'")

cnx.commit()  # noqa

# new attributes

NEW_ATTRIBUTES = (
    ("SEDADataObjectSystemId", "data_object_system_id"),
    ("SEDADataObjectGroupSystemId", "data_object_group_system_id"),
    ("SEDAClassificationAudience", "classification_audience"),
    ("SEDAEventTypeCode", "event_type_code"),
    ("SEDAOutcome", "outcome"),
    ("SEDAOutcomeDetail", "outcome_detail"),
    ("SEDAOutcomeDetailMessage", "outcome_detail_message"),
    ("SEDAEventDetailData", "event_detail_data"),
    ("SEDAIdentifier", "identifier"),
    ("SEDAGender", "gender"),
    ("SEDANationality", "nationality"),
    ("SEDAExternalReference", "external_reference"),
)

logger.info("\n\nAdd new attributes: \n")

for etype, attr in NEW_ATTRIBUTES:
    logger.info(f"Add new attribute '{attr}' for entity type '{etype}'")
    add_attribute(etype, attr)  # noqa

cnx.commit()  # noqa

# remove birth and death dates
drop_attribute("SEDABirthDate", "birth_date")
drop_attribute("SEDADeathDate", "death_date")

sync_schema_props_perms("SEDABirthDate")
sync_schema_props_perms("SEDADeathDate")
cnx.commit()


logger.info("\n\nCheck schema for seda 2.2 \n\n")

# check new subject relations have been added
CHECK_NEW_SUBJECT_RELATIONS = (
    (
        "seda_hold_rule_code_list_version_from",
        "SEDAHoldRuleCodeListVersion",
        "archive_transfer_hold_rule_code_list_version_from",
    ),
    (
        "seda_hold_rule_code_list_version_to",
        "SEDAHoldRuleCodeListVersion",
        "hold_rule_code_list_version_hold_rule_code_list_version_to",
    ),
    (
        "seda_data_object_profile",
        "SEDADataObjectProfile",
        "binary_data_object_data_object_profile",
    ),
    (
        "seda_hold_rule",
        "SEDAHoldRule",
        "archive_transfer_hold_rule",
    ),
    (
        "seda_seq_hold_rule_rule",
        "SEDAHoldRule",
        "hold_rule_seq_hold_rule_rule",
    ),
    (
        "seda_alt_hold_rule_prevent_inheritance",
        "SEDAHoldRule",
        "hold_rule_alt_hold_rule_prevent_inheritance",
    ),
    (
        "seda_linking_agent_identifier",
        "SEDALinkingAgentIdentifier",
        "event_linking_agent_identifier",
    ),
    (
        "seda_hold_rule",
        "SEDAHoldRule",
        "seq_alt_archive_unit_archive_unit_ref_id_management_hold_rule",
    ),
    (
        "seda_date_litteral",
        "SEDADateLitteral",
        "seq_alt_archive_unit_archive_unit_ref_id_management_date_litteral",
    ),
    (
        "seda_originating_system_id_reply_to",
        "SEDAOriginatingSystemIdReplyTo",
        "seq_alt_archive_unit_archive_unit_ref_id_management_originating_system_id_reply_to",
    ),
    (
        "seda_text_content",
        "SEDATextContent",
        "seq_alt_archive_unit_archive_unit_ref_id_management_text_content",
    ),
    (
        "seda_rule",
        "SEDASeqHoldRuleRule",
        "seq_hold_rule_rule_rule",
    ),
    (
        "seda_hold_end_date",
        "SEDAHoldEndDate",
        "seq_hold_rule_rule_hold_end_date",
    ),
    (
        "seda_hold_owner",
        "SEDAHoldOwner",
        "seq_hold_rule_rule_hold_owner",
    ),
    (
        "seda_hold_reassessing_date",
        "SEDAHoldReassessingDate",
        "seq_hold_rule_rule_hold_reassessing_date",
    ),
    (
        "seda_hold_reason",
        "SEDAHoldReason",
        "seq_hold_rule_rule_hold_reason",
    ),
    (
        "seda_prevent_rearrangement",
        "SEDAPreventRearrangement",
        "seq_hold_rule_rule_prevent_rearrangement",
    ),
    (
        "seda_linking_agent_identifier_type",
        "SEDALinkingAgentIdentifierType",
        "linking_agent_identifier_linking_agent_identifier_type",
    ),
    (
        "seda_linking_agent_identifier_value",
        "SEDALinkingAgentIdentifierValue",
        "linking_agent_identifier_linking_agent_identifier_value",
    ),
    (
        "seda_linking_agent_role",
        "SEDALinkingAgentRole",
        "linking_agent_identifier_linking_agent_role",
    ),
    (
        "seda_outcome_detail",  # old relation, still exists
        "SEDAOutcomeDetail",
        "event_outcome_detail",
    ),
    (
        "seda_agent",
        "SEDAAgent",
        "seq_alt_archive_unit_archive_unit_ref_id_management_agent",
    ),
)

for relname, obj, cw_relname in CHECK_NEW_SUBJECT_RELATIONS:
    eschema = cnx.repo.schema.entity_schema_for(obj)  # noqa
    try:
        eschema.relation_definition(relname)
        logger.info(f"Relation '{relname}' for '{obj}' exists")
    except KeyError:
        logger.error(f"Relation '{relname}' for '{obj} doesn't exist")


# check new object relations have been added
CHECK_NEW_OBJECT_RELATIONS = (
    (
        "seda_start_date",
        (
            "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement",
            "SEDASeqAccessRuleRule",
            "SEDASeqAppraisalRuleRule",
            "SEDASeqClassificationRuleRule",
            "SEDASeqDisseminationRuleRule",
            "SEDASeqHoldRuleRule",
            "SEDASeqReuseRuleRule",
            "SEDASeqStorageRuleRule",
        ),
        "seq_storage_rule_rule_start_date",
    ),
    (
        "seda_prevent_inheritance",
        (
            "SEDAAltAccessRulePreventInheritance",
            "SEDAAltAppraisalRulePreventInheritance",
            "SEDAAltClassificationRulePreventInheritance",
            "SEDAAltDisseminationRulePreventInheritance",
            "SEDAAltHoldRulePreventInheritance",
            "SEDAAltReuseRulePreventInheritance",
            "SEDAAltStorageRulePreventInheritance",
        ),
        "alt_storage_rule_prevent_inheritance_prevent_inheritance",
    ),
    (
        "seda_ref_non_rule_id_from",
        (
            "SEDAAltAccessRulePreventInheritance",
            "SEDAAltAppraisalRulePreventInheritance",
            "SEDAAltClassificationRulePreventInheritance",
            "SEDAAltDisseminationRulePreventInheritance",
            "SEDAAltHoldRulePreventInheritance",
            "SEDAAltReuseRulePreventInheritance",
            "SEDAAltStorageRulePreventInheritance",
        ),
        "alt_storage_rule_prevent_inheritance_ref_non_rule_id_from",
    ),
    (
        "seda_external_reference",
        (
            "SEDAAltIsPartOfArchiveUnitRefId",
            "SEDAAltIsVersionOfArchiveUnitRefId",
            "SEDAAltReferencesArchiveUnitRefId",
            "SEDAAltReplacesArchiveUnitRefId",
            "SEDAAltRequiresArchiveUnitRefId",
        ),
        "alt_is_version_of_archive_unit_ref_id_external_reference",
    ),
    (
        "seda_function",
        ("SEDAAgent",),
        "agent_function",
    ),
    (
        "seda_activity",
        ("SEDAAgent",),
        "agent_activity",
    ),
    (
        "seda_position",
        ("SEDAAgent",),
        "agent_position",
    ),
    (
        "seda_role",
        ("SEDAAgent",),
        "agent_role",
    ),
    (
        "seda_mandate",
        ("SEDAAgent",),
        "agent_mandate",
    ),
    (
        "seda_corpname",
        (
            "SEDAAltAgentCorpname",
            "SEDAAltSenderCorpname",
            "SEDAAltTransmitterCorpname",
        ),
        "alt_agent_corpname_corpname",
    ),
    (
        "seda_gender",
        (
            "SEDASeqAltAgentCorpnameFirstName",
            "SEDASeqAltSenderCorpnameFirstName",
            "SEDASeqAltTransmitterCorpnameFirstName",
        ),
        "seq_alt_agent_corpname_first_name_gender",
    ),
    (
        "seda_nationality",
        (
            "SEDASeqAltAgentCorpnameFirstName",
            "SEDASeqAltSenderCorpnameFirstName",
            "SEDASeqAltTransmitterCorpnameFirstName",
        ),
        "seq_alt_agent_corpname_first_name_nationality",
    ),
    (
        "seda_identifier",
        ("SEDAAgent", "SEDASender", "SEDATransmitter"),
        "agent_identifier",
    ),
    (
        "seda_first_name",
        (
            "SEDASeqAltAgentCorpnameFirstName",
            "SEDASeqAltSenderCorpnameFirstName",
            "SEDASeqAltTransmitterCorpnameFirstName",
        ),
        "seq_alt_agent_corpname_first_name_first_name",
    ),
    (
        "seda_full_name",
        (
            "SEDASeqAltAgentCorpnameFirstName",
            "SEDASeqAltSenderCorpnameFirstName",
            "SEDASeqAltTransmitterCorpnameFirstName",
        ),
        "seq_alt_agent_corpname_first_name_full_name",
    ),
    (
        "seda_given_name",
        (
            "SEDASeqAltAgentCorpnameFirstName",
            "SEDASeqAltSenderCorpnameFirstName",
            "SEDASeqAltTransmitterCorpnameFirstName",
        ),
        "seq_alt_agent_corpname_first_name_given_name",
    ),
    (
        "seda_birth_name",
        (
            "SEDASeqAltAgentCorpnameFirstName",
            "SEDASeqAltSenderCorpnameFirstName",
            "SEDASeqAltTransmitterCorpnameFirstName",
        ),
        "seq_alt_agent_corpname_first_name_birth_name",
    ),
    (
        "seda_birth_date",
        (
            "SEDASeqAltAgentCorpnameFirstName",
            "SEDASeqAltSenderCorpnameFirstName",
            "SEDASeqAltTransmitterCorpnameFirstName",
        ),
        "seq_alt_agent_corpname_first_name_birth_date",
    ),
    (
        "seda_birth_place",
        ("SEDASeqAltAgentCorpnameFirstName",),
        "seq_alt_agent_corpname_first_name_birth_place",
    ),
    (
        "seda_death_date",
        (
            "SEDASeqAltAgentCorpnameFirstName",
            "SEDASeqAltSenderCorpnameFirstName",
            "SEDASeqAltTransmitterCorpnameFirstName",
        ),
        "seq_alt_agent_corpname_first_name_death_date",
    ),
    (
        "seda_death_place",
        ("SEDASeqAltAgentCorpnameFirstName",),
        "seq_alt_agent_corpname_first_name_death_place",
    ),
    (
        "seda_alt_agent_corpname",
        ("SEDAAltAgentCorpname",),
        "agent_alt_agent_corpname",
    ),
    (
        "seda_seq_alt_agent_corpname_first_name",
        ("SEDASeqAltAgentCorpnameFirstName",),
        "alt_agent_corpname_seq_alt_agent_corpname_first_name",
    ),
)

for relname, objects, cw_relname in CHECK_NEW_OBJECT_RELATIONS:
    for obj in objects:
        eschema = cnx.repo.schema.entity_schema_for(obj)  # noqa
        try:
            eschema.relation_definition(relname, role="object")
            logger.info(f"Relation '{relname}' for '{obj}' exists")
        except KeyError:
            logger.error(f"ERROR, Relation '{relname}' for '{obj} doesn't exist")


# check new attributes have been added
for etype, attr in NEW_ATTRIBUTES:
    eschema = cnx.repo.schema.entity_schema_for(etype)  # noqa
    try:
        eschema.relation_definition(attr)
        logger.info("Attribute '%s' for '%s' is known" % (attr, etype))
    except KeyError:
        logger.error("ERROR, Attribute '%s' for '%s' doesn't exist" % (attr, etype))

CHECK_REMOVED_ATTRIBUTES = (
    ("SEDABirthDate", "birth_date"),
    ("SEDADeathDate", "death_date"),
)

for etype, attr in CHECK_REMOVED_ATTRIBUTES:
    eschema = cnx.repo.schema.entity_schema_for(etype)  # noqa
    try:
        eschema.relation_definition(attr)
        logger.error("ERROR, Attribute '%s' for '%s' is still known" % (attr, etype))
    except KeyError:
        logger.info("Attribute '%s' for '%s' doesn't exist" % (attr, etype))


# check add agent_abstract related relations have been removed
for rtype in (
    "seda_agent_abstract",
    "seda_alt_agent_abstract_corpname",
    "seda_seq_alt_agent_abstract_corpname_first_name",
):
    try:
        cnx.repo.schema.relation_schema_for(rtype)  # noqa
        logger.error("ERROR, Relation definition '%s' still exists" % rtype)
    except KeyError:
        logger.info("Relation definition '%s' doesn't exist" % rtype)

# relation en double (!!!)
# class event_linking_agent_identifier(RelationDefinition):

logger.info("\n\n Update compat_list for SEDA 2.1 : mark them SEDA 2.2 exportable \n")

SEDA22 = "SEDA 2.2"
COMPAT_SEP = ", "

for entity, compat in rql(
    "Any X, R WHERE X compat_list R"
).iter_rows_with_entities():  # noqa
    compats = compat.split(COMPAT_SEP)
    if "SEDA 2.1" in compats and not SEDA22 in compats:
        compats.append(SEDA22)
        entity.cw_set(compat_list=COMPAT_SEP.join(compats))

cnx.commit()  # noqa



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda unit tests for hooks"""

from os.path import dirname, join
from itertools import chain, repeat

from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_compound.entities import copy_entity
from cubicweb_seda import testutils, dataimport


class ValidationHooksTC(CubicWebTC):
    assertValidationError = testutils.assertValidationError

    def test_ref_non_rule_constraints(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            access_scheme = create("ConceptScheme", title="access")
            access_concept = access_scheme.add_concept(label="anyone")
            reuse_scheme = create("ConceptScheme", title="reuse")
            reuse_concept = reuse_scheme.add_concept(label="share-alike")
            cnx.commit()

            bdo = testutils.create_transfer_to_bdo(cnx)
            transfer = bdo.container[0]
            create(
                "SEDAAccessRuleCodeListVersion",
                seda_access_rule_code_list_version_from=transfer,
                seda_access_rule_code_list_version_to=access_scheme,
            )
            create(
                "SEDAReuseRuleCodeListVersion",
                seda_reuse_rule_code_list_version_from=transfer,
                seda_reuse_rule_code_list_version_to=reuse_scheme,
            )
            cnx.commit()

            rule_base = create("SEDAAccessRule", seda_access_rule=transfer)
            rule_alt = create(
                "SEDAAltAccessRulePreventInheritance",
                reverse_seda_alt_access_rule_prevent_inheritance=rule_base,
            )
            non_rule = create("SEDARefNonRuleId", seda_ref_non_rule_id_from=rule_alt)
            cnx.commit()

            with self.assertValidationError(cnx) as cm:
                non_rule.cw_set(seda_ref_non_rule_id_to=reuse_concept)
            self.assertIn("seda_ref_non_rule_id_to-subject", cm.exception.errors)

            non_rule.cw_set(seda_ref_non_rule_id_to=access_concept)
            cnx.commit()

    def test_empty_choice_created(self):
        """Check that a ValidationError is raised when an empty SEDAAlt... entity is created."""
        with self.admin_access.cnx() as cnx:
            # Create an empty SEDAAltAccessRulePreventInheritance
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            access_rule = cnx.create_entity("SEDAAccessRule", seda_access_rule=transfer)
            with self.assertValidationError(cnx) as cm:
                cnx.create_entity(
                    "SEDAAltAccessRulePreventInheritance",
                    reverse_seda_alt_access_rule_prevent_inheritance=access_rule,
                )
            self.assertIn("An alternative cannot be empty", str(cm.exception))

    def test_valid_choice_created(self):
        """Check that everything goes fine when a valid SEDAAlt... entity is created."""
        with self.admin_access.cnx() as cnx:
            # Create an empty SEDAArchiveUnit
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            access_rule = cnx.create_entity("SEDAAccessRule", seda_access_rule=transfer)
            choice = cnx.create_entity(
                "SEDAAltAccessRulePreventInheritance",
                reverse_seda_alt_access_rule_prevent_inheritance=access_rule,
            )
            cnx.create_entity(
                "SEDAPreventInheritance",
                prevent_inheritance=False,
                seda_prevent_inheritance=choice,
            )
            cnx.commit()

    def test_last_item_in_choice_deleted(self):
        """Check that a ValidationError is raised when the last relation from a SEDAAlt... entity
        is deleted.
        """
        with self.admin_access.cnx() as cnx:
            # Create a valid SEDAAltAccessRulePreventInheritance
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            access_rule = cnx.create_entity("SEDAAccessRule", seda_access_rule=transfer)
            choice = cnx.create_entity(
                "SEDAAltAccessRulePreventInheritance",
                reverse_seda_alt_access_rule_prevent_inheritance=access_rule,
            )
            rel = cnx.create_entity(
                "SEDAPreventInheritance",
                prevent_inheritance=False,
                seda_prevent_inheritance=choice,
            )
            cnx.commit()
            # Delete SEDAPreventInheritance
            with self.assertValidationError(cnx) as cm:
                cnx.execute(
                    "DELETE SEDAPreventInheritance X WHERE X eid %(rel_eid)s",
                    {"rel_eid": rel.eid},
                )
            self.assertIn("An alternative cannot be empty", str(cm.exception))

    def test_item_in_choice_deleted_with_remaining_item(self):
        """Check that everything goes fine when the a relation from a SEDAAlt... entity
        is deleted but another relation remains.
        """
        with self.admin_access.cnx() as cnx:
            # Create a valid SEDAAltAccessRulePreventInheritance with two items
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            access_rule = cnx.create_entity("SEDAAccessRule", seda_access_rule=transfer)
            choice = cnx.create_entity(
                "SEDAAltAccessRulePreventInheritance",
                reverse_seda_alt_access_rule_prevent_inheritance=access_rule,
            )
            rel = cnx.create_entity(
                "SEDAPreventInheritance",
                prevent_inheritance=False,
                seda_prevent_inheritance=choice,
            )
            cnx.create_entity("SEDARefNonRuleId", seda_ref_non_rule_id_from=choice)
            cnx.commit()
            # Delete SEDAPreventInheritance
            cnx.execute(
                "DELETE SEDAPreventInheritance X WHERE X eid %(rel_eid)s",
                {"rel_eid": rel.eid},
            )
            cnx.commit()


class SetDefaultHooksTC(CubicWebTC):
    def test_default_code_list_version(self):
        with self.admin_access.cnx() as cnx:
            for rtype, etype in chain(
                zip(("file_category", "seda_encoding_to"), repeat(None)),
                [
                    ("seda_algorithm", "SEDABinaryDataObject"),
                    ("seda_rule", "SEDASeqAppraisalRuleRule"),
                    ("seda_rule", "SEDASeqAccessRuleRule"),
                    ("seda_rule", "SEDASeqDisseminationRuleRule"),
                ],
            ):
                testutils.scheme_for_type(cnx, rtype, etype)
            cnx.commit()
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            cnx.commit()

            self.assertTrue(transfer.reverse_seda_file_format_code_list_version_from)
            self.assertTrue(
                transfer.reverse_seda_message_digest_algorithm_code_list_version_from
            )
            self.assertTrue(transfer.reverse_seda_mime_type_code_list_version_from)
            self.assertTrue(transfer.reverse_seda_encoding_code_list_version_from)
            self.assertTrue(transfer.reverse_seda_access_rule_code_list_version_from)
            self.assertTrue(transfer.reverse_seda_appraisal_rule_code_list_version_from)
            self.assertTrue(
                transfer.reverse_seda_dissemination_rule_code_list_version_from
            )

    def test_default_card_on_typed_data_object_ref(self):
        """When creating a SEDADataObjectReference in the context of a reference, its cardinality
        should always be 1 (by default any card is allowed since in may be used in the context of
        'main' reference).
        """
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            version_of = cnx.create_entity(
                "SEDAIsVersionOf", seda_is_version_of=unit_alt_seq
            )
            alt2 = cnx.create_entity(
                "SEDAAltIsVersionOfArchiveUnitRefId",
                reverse_seda_alt_is_version_of_archive_unit_ref_id=version_of,
            )
            do_ref = cnx.create_entity(
                "SEDADataObjectReference", seda_data_object_reference=alt2
            )
            cnx.commit()

            self.assertEqual(do_ref.user_cardinality, "1")

    def test_simplified_profile_do_ref_sync(self):
        """When creating a SEDABinaryDataObject in a simplified profile from the ui, one only
        specifies user_cardinality for it, not for it's associated reference whose cardinality
        should have the same value.
        """
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity(
                "SEDAArchiveTransfer", title="test profile", simplified_profile=True
            )
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            bdo = testutils.create_data_object(
                unit_alt_seq, user_cardinality="1", seda_binary_data_object=transfer
            )
            cnx.commit()
            ref = cnx.find("SEDADataObjectReference").one()

            ref.cw_clear_all_caches()
            self.assertEqual(ref.user_cardinality, "1")

            bdo.cw_set(user_cardinality="1..n")
            cnx.commit()
            ref.cw_clear_all_caches()
            self.assertEqual(ref.user_cardinality, "1..n")


class CheckProfileTC(CubicWebTC):
    assertValidationError = testutils.assertValidationError

    def test_base(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity(
                "SEDAArchiveTransfer", title="diagnosis testing"
            )
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            access_rule = cnx.create_entity(
                "SEDAAccessRule", seda_access_rule=unit_alt_seq
            )
            cnx.commit()

            with self.assertValidationError(cnx):
                transfer.cw_set(simplified_profile=True)

            access_rule_seq = cnx.create_entity(
                "SEDASeqAccessRuleRule", reverse_seda_seq_access_rule_rule=access_rule
            )
            start_date = cnx.create_entity(
                "SEDAStartDate", seda_start_date=access_rule_seq
            )
            transfer.cw_set(simplified_profile=True)
            cnx.commit()

            with self.assertValidationError(cnx):
                start_date.cw_set(user_cardinality="0..1")

            with self.assertValidationError(cnx):
                access_rule_seq.cw_delete()


class DispatchFileCategoryTC(CubicWebTC):
    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            dataimport.import_seda_schemes(
                cnx,
                lcsv_files=[
                    (
                        "Categories de fichier",
                        "file_category",
                        (),
                        join(dirname(__file__), "data", "file_categories.csv"),
                    )
                ],
            )
            self.categories = dict(
                cnx.execute(
                    "Any LL, C WHERE L label_of C, L label LL, "
                    'C in_scheme CS, CS title "Categories de fichier"'
                )
            )
            transfer = cnx.create_entity(
                "SEDAArchiveTransfer", title="test profile", simplified_profile=True
            )
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            bdo = testutils.create_data_object(
                unit_alt_seq, seda_binary_data_object=transfer
            )
            cnx.commit()

        self.transfer_eid = transfer.eid
        self.unit_alt_seq_eid = unit_alt_seq.eid
        self.bdo_eid = bdo.eid

    def test_base(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            unit_alt_seq = cnx.entity_from_eid(self.unit_alt_seq_eid)
            bdo = cnx.entity_from_eid(self.bdo_eid)

            bdo.cw_set(file_category=self.categories["document"])
            cnx.commit()
            self.assertFormatEqual(
                bdo,
                ["application/msword", "application/pdf"],
                ["fmt/37", "fmt/38", "fmt/14"],
            )

            bdo.cw_set(file_category=None)
            cnx.commit()
            self.assertFormatEqual(bdo, [], [])

            bdo.cw_set(file_category=self.categories["doc"])
            cnx.commit()
            self.assertFormatEqual(bdo, ["application/msword"], ["fmt/37", "fmt/38"])

            # test no dispatch occurs on component archive unit...
            unit_comp, unit_comp_alt, unit_comp_alt_seq = testutils.create_archive_unit(
                None, cnx=cnx
            )
            bdo_comp = testutils.create_data_object(unit_comp_alt_seq)
            cnx.commit()

            bdo_comp.cw_set(file_category=self.categories["document"])
            cnx.commit()
            self.assertFormatEqual(bdo_comp, [], [])
            # until it's imported into some parent
            copy_entity(unit_comp, seda_archive_unit=unit_alt_seq, clone_of=unit_comp)
            cnx.commit()
            (imported_bdo,) = (
                x for x in transfer.binary_data_objects if x.eid != bdo.eid
            )
            self.assertFormatEqual(
                imported_bdo,
                ["application/msword", "application/pdf"],
                ["fmt/37", "fmt/38", "fmt/14"],
            )

            # ensure deleting the data object doesn't raise an exception because
            # of erroneous hook
            bdo.cw_delete()
            cnx.commit()

    def test_join(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            bdo = cnx.entity_from_eid(self.bdo_eid)

            formats = testutils.scheme_for_type(
                cnx, "seda_format_id_to", None, "fmt/37", "fmt/14"
            )
            mime_types = testutils.scheme_for_type(
                cnx, "seda_mime_type_to", None, "application/msword", "application/pdf"
            )
            transfer.reverse_seda_file_format_code_list_version_from[0].cw_set(
                seda_file_format_code_list_version_to=formats
            )
            transfer.reverse_seda_mime_type_code_list_version_from[0].cw_set(
                seda_mime_type_code_list_version_to=mime_types
            )
            cnx.commit()

            bdo.cw_set(file_category=self.categories["document"])
            cnx.commit()
            self.assertFormatEqual(
                bdo, ["application/msword", "application/pdf"], ["fmt/37", "fmt/14"]
            )

    def assertFormatEqual(self, bdo, expected_mime_types, expected_format_ids):
        bdo.cw_clear_all_caches()
        bdo.reverse_seda_mime_type_from[0].cw_clear_all_caches()
        bdo.reverse_seda_format_id_from[0].cw_clear_all_caches()
        mime_types = {
            x.label() for x in bdo.reverse_seda_mime_type_from[0].seda_mime_type_to
        }
        self.assertEqual(mime_types, set(expected_mime_types))
        format_ids = {
            x.label() for x in bdo.reverse_seda_format_id_from[0].seda_format_id_to
        }
        self.assertEqual(format_ids, set(expected_format_ids))


class OrderingHooksTC(CubicWebTC):
    def test_first_level_archive_units(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            unit1 = testutils.create_archive_unit(transfer)[0]
            unit2 = testutils.create_archive_unit(transfer)[0]
            cnx.commit()

            self.assertEqual(unit1.ordering, 1)
            self.assertEqual(unit2.ordering, 2)

    def test_archive_units(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            unit1 = testutils.create_archive_unit(unit_alt_seq)[0]
            unit2 = testutils.create_archive_unit(unit_alt_seq)[0]
            cnx.commit()

            self.assertEqual(unit1.ordering, 1)
            self.assertEqual(unit2.ordering, 2)

    def test_data_objects(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity(
                "SEDAArchiveTransfer", title="test profile", simplified_profile=True
            )
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            bdo1 = testutils.create_data_object(
                unit_alt_seq, user_cardinality="1", seda_binary_data_object=transfer
            )
            bdo2 = testutils.create_data_object(
                unit_alt_seq, user_cardinality="1", seda_binary_data_object=transfer
            )
            cnx.commit()

            self.assertEqual(bdo1.ordering, 1)
            self.assertEqual(bdo2.ordering, 2)
            self.assertEqual(bdo1.reverse_seda_data_object_reference_id[0].ordering, 1)
            self.assertEqual(bdo2.reverse_seda_data_object_reference_id[0].ordering, 2)

    def test_remove_keep_in_sync(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity("SEDAArchiveTransfer", title="test profile")
            bdo1 = testutils.create_data_object(transfer)
            bdo2 = testutils.create_data_object(transfer)
            cnx.commit()

            bdo1.cw_delete()
            cnx.commit()

            bdo2.cw_clear_all_caches()
            self.assertEqual(bdo2.ordering, 1)


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""THIS FILE IS GENERATED FROM SEDA 2.2 XSD FILES, DO NOT EDIT"""

from cubicweb.entities import AnyEntity, fetch_config


class SEDAAnyEntity(AnyEntity):
    __abstract__ = True
    value_attr = None

    def dc_title(self):
        if self.value_attr is None:
            return self.dc_type()
        return self.printable_value(self.value_attr)


class SEDAArchiveTransfer(SEDAAnyEntity):
    __regid__ = "SEDAArchiveTransfer"
    fetch_attrs, cw_fetch_order = fetch_config(["title", "user_annotation"])
    value_attr = None


class SEDAComment(SEDAAnyEntity):
    __regid__ = "SEDAComment"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "comment", "user_annotation"]
    )
    value_attr = "comment"


class SEDASignature(SEDAAnyEntity):
    __regid__ = "SEDASignature"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAArchivalAgreement(SEDAAnyEntity):
    __regid__ = "SEDAArchivalAgreement"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "archival_agreement", "user_annotation"]
    )
    value_attr = "archival_agreement"


class SEDARelatedTransferReference(SEDAAnyEntity):
    __regid__ = "SEDARelatedTransferReference"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDATransferRequestReplyIdentifier(SEDAAnyEntity):
    __regid__ = "SEDATransferRequestReplyIdentifier"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAMessageDigestAlgorithmCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDAMessageDigestAlgorithmCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAMimeTypeCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDAMimeTypeCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAEncodingCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDAEncodingCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAFileFormatCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDAFileFormatCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDACompressionAlgorithmCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDACompressionAlgorithmCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADataObjectVersionCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDADataObjectVersionCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAStorageRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDAStorageRuleCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAppraisalRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDAAppraisalRuleCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAccessRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDAAccessRuleCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADisseminationRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDADisseminationRuleCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAReuseRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDAReuseRuleCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAClassificationRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDAClassificationRuleCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAHoldRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDAHoldRuleCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAcquisitionInformationCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDAAcquisitionInformationCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDARelationshipCodeListVersion(SEDAAnyEntity):
    __regid__ = "SEDARelationshipCodeListVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDABinaryDataObject(SEDAAnyEntity):
    __regid__ = "SEDABinaryDataObject"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["ordering", "filename", "user_cardinality", "user_annotation"]
    )
    value_attr = None


class SEDADataObjectGroup(SEDAAnyEntity):
    __regid__ = "SEDADataObjectGroup"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAPhysicalDataObject(SEDAAnyEntity):
    __regid__ = "SEDAPhysicalDataObject"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["ordering", "user_cardinality", "user_annotation"]
    )
    value_attr = None


class SEDADataObjectProfile(SEDAAnyEntity):
    __regid__ = "SEDADataObjectProfile"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "data_object_profile", "user_annotation"]
    )
    value_attr = "data_object_profile"


class SEDADataObjectSystemId(SEDAAnyEntity):
    __regid__ = "SEDADataObjectSystemId"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "data_object_system_id", "user_annotation"]
    )
    value_attr = "data_object_system_id"


class SEDADataObjectGroupSystemId(SEDAAnyEntity):
    __regid__ = "SEDADataObjectGroupSystemId"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "data_object_group_system_id", "user_annotation"]
    )
    value_attr = "data_object_group_system_id"


class SEDARelationship(SEDAAnyEntity):
    __regid__ = "SEDARelationship"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDADataObjectVersion(SEDAAnyEntity):
    __regid__ = "SEDADataObjectVersion"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAltBinaryDataObjectAttachment(SEDAAnyEntity):
    __regid__ = "SEDAAltBinaryDataObjectAttachment"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDASize(SEDAAnyEntity):
    __regid__ = "SEDASize"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDACompressed(SEDAAnyEntity):
    __regid__ = "SEDACompressed"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "compressed", "user_annotation"]
    )
    value_attr = "compressed"


class SEDALogBook(SEDAAnyEntity):
    __regid__ = "SEDALogBook"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAPhysicalId(SEDAAnyEntity):
    __regid__ = "SEDAPhysicalId"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAArchiveUnit(SEDAAnyEntity):
    __regid__ = "SEDAArchiveUnit"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["ordering", "user_cardinality", "user_annotation"]
    )
    value_attr = None


class SEDAArchivalProfile(SEDAAnyEntity):
    __regid__ = "SEDAArchivalProfile"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "archival_profile", "user_annotation"]
    )
    value_attr = "archival_profile"


class SEDAServiceLevel(SEDAAnyEntity):
    __regid__ = "SEDAServiceLevel"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "service_level", "user_annotation"]
    )
    value_attr = "service_level"


class SEDAAcquisitionInformation(SEDAAnyEntity):
    __regid__ = "SEDAAcquisitionInformation"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDALegalStatus(SEDAAnyEntity):
    __regid__ = "SEDALegalStatus"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAOriginatingAgencyIdentifier(SEDAAnyEntity):
    __regid__ = "SEDAOriginatingAgencyIdentifier"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "originating_agency_identifier", "user_annotation"]
    )
    value_attr = "originating_agency_identifier"


class SEDASubmissionAgencyIdentifier(SEDAAnyEntity):
    __regid__ = "SEDASubmissionAgencyIdentifier"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "submission_agency_identifier", "user_annotation"]
    )
    value_attr = "submission_agency_identifier"


class SEDAStorageRule(SEDAAnyEntity):
    __regid__ = "SEDAStorageRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAppraisalRule(SEDAAnyEntity):
    __regid__ = "SEDAAppraisalRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAccessRule(SEDAAnyEntity):
    __regid__ = "SEDAAccessRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADisseminationRule(SEDAAnyEntity):
    __regid__ = "SEDADisseminationRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAReuseRule(SEDAAnyEntity):
    __regid__ = "SEDAReuseRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAClassificationRule(SEDAAnyEntity):
    __regid__ = "SEDAClassificationRule"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "classification_owner", "user_annotation"]
    )
    value_attr = "classification_owner"


class SEDANeedAuthorization(SEDAAnyEntity):
    __regid__ = "SEDANeedAuthorization"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "need_authorization", "user_annotation"]
    )
    value_attr = "need_authorization"


class SEDAHoldRule(SEDAAnyEntity):
    __regid__ = "SEDAHoldRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAltArchiveUnitArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = "SEDAAltArchiveUnitArchiveUnitRefId"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAAttachment(SEDAAnyEntity):
    __regid__ = "SEDAAttachment"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAUri(SEDAAnyEntity):
    __regid__ = "SEDAUri"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAFormatLitteral(SEDAAnyEntity):
    __regid__ = "SEDAFormatLitteral"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "format_litteral", "user_annotation"]
    )
    value_attr = "format_litteral"


class SEDAMimeType(SEDAAnyEntity):
    __regid__ = "SEDAMimeType"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAFormatId(SEDAAnyEntity):
    __regid__ = "SEDAFormatId"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAEncoding(SEDAAnyEntity):
    __regid__ = "SEDAEncoding"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDACreatingApplicationName(SEDAAnyEntity):
    __regid__ = "SEDACreatingApplicationName"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "creating_application_name", "user_annotation"]
    )
    value_attr = "creating_application_name"


class SEDACreatingApplicationVersion(SEDAAnyEntity):
    __regid__ = "SEDACreatingApplicationVersion"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "creating_application_version", "user_annotation"]
    )
    value_attr = "creating_application_version"


class SEDADateCreatedByApplication(SEDAAnyEntity):
    __regid__ = "SEDADateCreatedByApplication"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDACreatingOs(SEDAAnyEntity):
    __regid__ = "SEDACreatingOs"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "creating_os", "user_annotation"]
    )
    value_attr = "creating_os"


class SEDACreatingOsVersion(SEDAAnyEntity):
    __regid__ = "SEDACreatingOsVersion"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "creating_os_version", "user_annotation"]
    )
    value_attr = "creating_os_version"


class SEDALastModified(SEDAAnyEntity):
    __regid__ = "SEDALastModified"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAEvent(SEDAAnyEntity):
    __regid__ = "SEDAEvent"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDAWidth(SEDAAnyEntity):
    __regid__ = "SEDAWidth"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAHeight(SEDAAnyEntity):
    __regid__ = "SEDAHeight"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADepth(SEDAAnyEntity):
    __regid__ = "SEDADepth"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAShape(SEDAAnyEntity):
    __regid__ = "SEDAShape"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADiameter(SEDAAnyEntity):
    __regid__ = "SEDADiameter"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDALength(SEDAAnyEntity):
    __regid__ = "SEDALength"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAThickness(SEDAAnyEntity):
    __regid__ = "SEDAThickness"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAWeight(SEDAAnyEntity):
    __regid__ = "SEDAWeight"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDANumberOfPage(SEDAAnyEntity):
    __regid__ = "SEDANumberOfPage"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = "SEDAArchiveUnitRefId"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDASeqAltArchiveUnitArchiveUnitRefIdManagement(SEDAAnyEntity):
    __regid__ = "SEDASeqAltArchiveUnitArchiveUnitRefIdManagement"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDASeqStorageRuleRule(SEDAAnyEntity):
    __regid__ = "SEDASeqStorageRuleRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAltStorageRulePreventInheritance(SEDAAnyEntity):
    __regid__ = "SEDAAltStorageRulePreventInheritance"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDASeqAppraisalRuleRule(SEDAAnyEntity):
    __regid__ = "SEDASeqAppraisalRuleRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAltAppraisalRulePreventInheritance(SEDAAnyEntity):
    __regid__ = "SEDAAltAppraisalRulePreventInheritance"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDASeqAccessRuleRule(SEDAAnyEntity):
    __regid__ = "SEDASeqAccessRuleRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAltAccessRulePreventInheritance(SEDAAnyEntity):
    __regid__ = "SEDAAltAccessRulePreventInheritance"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDASeqDisseminationRuleRule(SEDAAnyEntity):
    __regid__ = "SEDASeqDisseminationRuleRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAltDisseminationRulePreventInheritance(SEDAAnyEntity):
    __regid__ = "SEDAAltDisseminationRulePreventInheritance"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDASeqReuseRuleRule(SEDAAnyEntity):
    __regid__ = "SEDASeqReuseRuleRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAltReuseRulePreventInheritance(SEDAAnyEntity):
    __regid__ = "SEDAAltReuseRulePreventInheritance"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDASeqClassificationRuleRule(SEDAAnyEntity):
    __regid__ = "SEDASeqClassificationRuleRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAClassificationAudience(SEDAAnyEntity):
    __regid__ = "SEDAClassificationAudience"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "classification_audience", "user_annotation"]
    )
    value_attr = "classification_audience"


class SEDAAltClassificationRulePreventInheritance(SEDAAnyEntity):
    __regid__ = "SEDAAltClassificationRulePreventInheritance"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAClassificationReassessingDate(SEDAAnyEntity):
    __regid__ = "SEDAClassificationReassessingDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDANeedReassessingAuthorization(SEDAAnyEntity):
    __regid__ = "SEDANeedReassessingAuthorization"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "need_reassessing_authorization", "user_annotation"]
    )
    value_attr = "need_reassessing_authorization"


class SEDASeqHoldRuleRule(SEDAAnyEntity):
    __regid__ = "SEDASeqHoldRuleRule"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAltHoldRulePreventInheritance(SEDAAnyEntity):
    __regid__ = "SEDAAltHoldRulePreventInheritance"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAEventIdentifier(SEDAAnyEntity):
    __regid__ = "SEDAEventIdentifier"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAEventTypeCode(SEDAAnyEntity):
    __regid__ = "SEDAEventTypeCode"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "event_type_code", "user_annotation"]
    )
    value_attr = "event_type_code"


class SEDAEventType(SEDAAnyEntity):
    __regid__ = "SEDAEventType"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAEventDetail(SEDAAnyEntity):
    __regid__ = "SEDAEventDetail"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAOutcome(SEDAAnyEntity):
    __regid__ = "SEDAOutcome"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "outcome", "user_annotation"]
    )
    value_attr = "outcome"


class SEDAOutcomeDetail(SEDAAnyEntity):
    __regid__ = "SEDAOutcomeDetail"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "outcome_detail", "user_annotation"]
    )
    value_attr = "outcome_detail"


class SEDAOutcomeDetailMessage(SEDAAnyEntity):
    __regid__ = "SEDAOutcomeDetailMessage"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "outcome_detail_message", "user_annotation"]
    )
    value_attr = "outcome_detail_message"


class SEDAEventDetailData(SEDAAnyEntity):
    __regid__ = "SEDAEventDetailData"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "event_detail_data", "user_annotation"]
    )
    value_attr = "event_detail_data"


class SEDALinkingAgentIdentifier(SEDAAnyEntity):
    __regid__ = "SEDALinkingAgentIdentifier"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAEventAbstract(SEDAAnyEntity):
    __regid__ = "SEDAEventAbstract"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADataObjectReferenceId(SEDAAnyEntity):
    __regid__ = "SEDADataObjectReferenceId"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADataObjectReference(SEDAAnyEntity):
    __regid__ = "SEDADataObjectReference"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDADescriptionLevel(SEDAAnyEntity):
    __regid__ = "SEDADescriptionLevel"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDATitle(SEDAAnyEntity):
    __regid__ = "SEDATitle"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "title", "user_annotation"]
    )
    value_attr = "title"


class SEDAFilePlanPosition(SEDAAnyEntity):
    __regid__ = "SEDAFilePlanPosition"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "file_plan_position", "user_annotation"]
    )
    value_attr = "file_plan_position"


class SEDASystemId(SEDAAnyEntity):
    __regid__ = "SEDASystemId"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAOriginatingSystemId(SEDAAnyEntity):
    __regid__ = "SEDAOriginatingSystemId"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAArchivalAgencyArchiveUnitIdentifier(SEDAAnyEntity):
    __regid__ = "SEDAArchivalAgencyArchiveUnitIdentifier"
    fetch_attrs, cw_fetch_order = fetch_config(
        [
            "user_cardinality",
            "archival_agency_archive_unit_identifier",
            "user_annotation",
        ]
    )
    value_attr = "archival_agency_archive_unit_identifier"


class SEDAOriginatingAgencyArchiveUnitIdentifier(SEDAAnyEntity):
    __regid__ = "SEDAOriginatingAgencyArchiveUnitIdentifier"
    fetch_attrs, cw_fetch_order = fetch_config(
        [
            "user_cardinality",
            "originating_agency_archive_unit_identifier",
            "user_annotation",
        ]
    )
    value_attr = "originating_agency_archive_unit_identifier"


class SEDATransferringAgencyArchiveUnitIdentifier(SEDAAnyEntity):
    __regid__ = "SEDATransferringAgencyArchiveUnitIdentifier"
    fetch_attrs, cw_fetch_order = fetch_config(
        [
            "user_cardinality",
            "transferring_agency_archive_unit_identifier",
            "user_annotation",
        ]
    )
    value_attr = "transferring_agency_archive_unit_identifier"


class SEDADescription(SEDAAnyEntity):
    __regid__ = "SEDADescription"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "description", "user_annotation"]
    )
    value_attr = "description"


class SEDAType(SEDAAnyEntity):
    __regid__ = "SEDAType"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADocumentType(SEDAAnyEntity):
    __regid__ = "SEDADocumentType"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "document_type", "user_annotation"]
    )
    value_attr = "document_type"


class SEDALanguage(SEDAAnyEntity):
    __regid__ = "SEDALanguage"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADescriptionLanguage(SEDAAnyEntity):
    __regid__ = "SEDADescriptionLanguage"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAStatus(SEDAAnyEntity):
    __regid__ = "SEDAStatus"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "status", "user_annotation"]
    )
    value_attr = "status"


class SEDAVersion(SEDAAnyEntity):
    __regid__ = "SEDAVersion"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "version", "user_annotation"]
    )
    value_attr = "version"


class SEDATag(SEDAAnyEntity):
    __regid__ = "SEDATag"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["ordering", "user_cardinality", "tag", "user_annotation"]
    )
    value_attr = "tag"


class SEDAKeyword(SEDAAnyEntity):
    __regid__ = "SEDAKeyword"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["ordering", "user_cardinality", "keyword_content", "user_annotation"]
    )
    value_attr = "keyword_content"


class SEDAOriginatingAgency(SEDAAnyEntity):
    __regid__ = "SEDAOriginatingAgency"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDASubmissionAgency(SEDAAnyEntity):
    __regid__ = "SEDASubmissionAgency"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAgent(SEDAAnyEntity):
    __regid__ = "SEDAAgent"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAuthorizedAgent(SEDAAnyEntity):
    __regid__ = "SEDAAuthorizedAgent"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAWriter(SEDAAnyEntity):
    __regid__ = "SEDAWriter"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDAAddressee(SEDAAnyEntity):
    __regid__ = "SEDAAddressee"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDARecipient(SEDAAnyEntity):
    __regid__ = "SEDARecipient"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDATransmitter(SEDAAnyEntity):
    __regid__ = "SEDATransmitter"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDASender(SEDAAnyEntity):
    __regid__ = "SEDASender"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDASource(SEDAAnyEntity):
    __regid__ = "SEDASource"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "source", "user_annotation"]
    )
    value_attr = "source"


class SEDACreatedDate(SEDAAnyEntity):
    __regid__ = "SEDACreatedDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDATransactedDate(SEDAAnyEntity):
    __regid__ = "SEDATransactedDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAAcquiredDate(SEDAAnyEntity):
    __regid__ = "SEDAAcquiredDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDASentDate(SEDAAnyEntity):
    __regid__ = "SEDASentDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAReceivedDate(SEDAAnyEntity):
    __regid__ = "SEDAReceivedDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDARegisteredDate(SEDAAnyEntity):
    __regid__ = "SEDARegisteredDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAStartDate(SEDAAnyEntity):
    __regid__ = "SEDAStartDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAEndDate(SEDAAnyEntity):
    __regid__ = "SEDAEndDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADateLitteral(SEDAAnyEntity):
    __regid__ = "SEDADateLitteral"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "date_litteral", "user_annotation"]
    )
    value_attr = "date_litteral"


class SEDAOriginatingSystemIdReplyTo(SEDAAnyEntity):
    __regid__ = "SEDAOriginatingSystemIdReplyTo"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "originating_system_id_reply_to", "user_annotation"]
    )
    value_attr = "originating_system_id_reply_to"


class SEDATextContent(SEDAAnyEntity):
    __regid__ = "SEDATextContent"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "text_content", "user_annotation"]
    )
    value_attr = "text_content"


class SEDAPreventInheritance(SEDAAnyEntity):
    __regid__ = "SEDAPreventInheritance"
    fetch_attrs, cw_fetch_order = fetch_config(["prevent_inheritance"])
    value_attr = "prevent_inheritance"


class SEDARefNonRuleId(SEDAAnyEntity):
    __regid__ = "SEDARefNonRuleId"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAHoldEndDate(SEDAAnyEntity):
    __regid__ = "SEDAHoldEndDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAHoldOwner(SEDAAnyEntity):
    __regid__ = "SEDAHoldOwner"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "hold_owner", "user_annotation"]
    )
    value_attr = "hold_owner"


class SEDAHoldReassessingDate(SEDAAnyEntity):
    __regid__ = "SEDAHoldReassessingDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAHoldReason(SEDAAnyEntity):
    __regid__ = "SEDAHoldReason"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "hold_reason", "user_annotation"]
    )
    value_attr = "hold_reason"


class SEDAPreventRearrangement(SEDAAnyEntity):
    __regid__ = "SEDAPreventRearrangement"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "prevent_rearrangement", "user_annotation"]
    )
    value_attr = "prevent_rearrangement"


class SEDALinkingAgentIdentifierType(SEDAAnyEntity):
    __regid__ = "SEDALinkingAgentIdentifierType"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "linking_agent_identifier_type", "user_annotation"]
    )
    value_attr = "linking_agent_identifier_type"


class SEDALinkingAgentIdentifierValue(SEDAAnyEntity):
    __regid__ = "SEDALinkingAgentIdentifierValue"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "linking_agent_identifier_value", "user_annotation"]
    )
    value_attr = "linking_agent_identifier_value"


class SEDALinkingAgentRole(SEDAAnyEntity):
    __regid__ = "SEDALinkingAgentRole"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "linking_agent_role", "user_annotation"]
    )
    value_attr = "linking_agent_role"


class SEDACustodialHistoryItem(SEDAAnyEntity):
    __regid__ = "SEDACustodialHistoryItem"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["ordering", "user_cardinality", "custodial_history_item", "user_annotation"]
    )
    value_attr = "custodial_history_item"


class SEDACustodialHistoryFile(SEDAAnyEntity):
    __regid__ = "SEDACustodialHistoryFile"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAKeywordReference(SEDAAnyEntity):
    __regid__ = "SEDAKeywordReference"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAKeywordType(SEDAAnyEntity):
    __regid__ = "SEDAKeywordType"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDASpatial(SEDAAnyEntity):
    __regid__ = "SEDASpatial"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["ordering", "user_cardinality", "spatial", "user_annotation"]
    )
    value_attr = "spatial"


class SEDATemporal(SEDAAnyEntity):
    __regid__ = "SEDATemporal"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["ordering", "user_cardinality", "temporal", "user_annotation"]
    )
    value_attr = "temporal"


class SEDAJuridictional(SEDAAnyEntity):
    __regid__ = "SEDAJuridictional"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["ordering", "user_cardinality", "juridictional", "user_annotation"]
    )
    value_attr = "juridictional"


class SEDAAltAgentCorpname(SEDAAnyEntity):
    __regid__ = "SEDAAltAgentCorpname"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAIdentifier(SEDAAnyEntity):
    __regid__ = "SEDAIdentifier"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "identifier", "user_annotation"]
    )
    value_attr = "identifier"


class SEDAFunction(SEDAAnyEntity):
    __regid__ = "SEDAFunction"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "function", "user_annotation"]
    )
    value_attr = "function"


class SEDAActivity(SEDAAnyEntity):
    __regid__ = "SEDAActivity"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "activity", "user_annotation"]
    )
    value_attr = "activity"


class SEDAPosition(SEDAAnyEntity):
    __regid__ = "SEDAPosition"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "position", "user_annotation"]
    )
    value_attr = "position"


class SEDARole(SEDAAnyEntity):
    __regid__ = "SEDARole"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "role", "user_annotation"]
    )
    value_attr = "role"


class SEDAMandate(SEDAAnyEntity):
    __regid__ = "SEDAMandate"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "mandate", "user_annotation"]
    )
    value_attr = "mandate"


class SEDAAltTransmitterCorpname(SEDAAnyEntity):
    __regid__ = "SEDAAltTransmitterCorpname"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAAltSenderCorpname(SEDAAnyEntity):
    __regid__ = "SEDAAltSenderCorpname"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAIsVersionOf(SEDAAnyEntity):
    __regid__ = "SEDAIsVersionOf"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDAReplaces(SEDAAnyEntity):
    __regid__ = "SEDAReplaces"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDARequires(SEDAAnyEntity):
    __regid__ = "SEDARequires"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDAIsPartOf(SEDAAnyEntity):
    __regid__ = "SEDAIsPartOf"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDAReferences(SEDAAnyEntity):
    __regid__ = "SEDAReferences"
    fetch_attrs, cw_fetch_order = fetch_config(["ordering", "user_cardinality"])
    value_attr = None


class SEDAGpsVersionID(SEDAAnyEntity):
    __regid__ = "SEDAGpsVersionID"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAGpsAltitude(SEDAAnyEntity):
    __regid__ = "SEDAGpsAltitude"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAGpsAltitudeRef(SEDAAnyEntity):
    __regid__ = "SEDAGpsAltitudeRef"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAGpsLatitude(SEDAAnyEntity):
    __regid__ = "SEDAGpsLatitude"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAGpsLatitudeRef(SEDAAnyEntity):
    __regid__ = "SEDAGpsLatitudeRef"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAGpsLongitude(SEDAAnyEntity):
    __regid__ = "SEDAGpsLongitude"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAGpsLongitudeRef(SEDAAnyEntity):
    __regid__ = "SEDAGpsLongitudeRef"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAGpsDateStamp(SEDAAnyEntity):
    __regid__ = "SEDAGpsDateStamp"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDAwhen(SEDAAnyEntity):
    __regid__ = "SEDAwhen"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDACorpname(SEDAAnyEntity):
    __regid__ = "SEDACorpname"
    fetch_attrs, cw_fetch_order = fetch_config(["corpname"])
    value_attr = "corpname"


class SEDASeqAltAgentCorpnameFirstName(SEDAAnyEntity):
    __regid__ = "SEDASeqAltAgentCorpnameFirstName"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDASeqAltTransmitterCorpnameFirstName(SEDAAnyEntity):
    __regid__ = "SEDASeqAltTransmitterCorpnameFirstName"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDASeqAltSenderCorpnameFirstName(SEDAAnyEntity):
    __regid__ = "SEDASeqAltSenderCorpnameFirstName"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAAltIsVersionOfArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = "SEDAAltIsVersionOfArchiveUnitRefId"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAAltReplacesArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = "SEDAAltReplacesArchiveUnitRefId"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAAltRequiresArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = "SEDAAltRequiresArchiveUnitRefId"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAAltIsPartOfArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = "SEDAAltIsPartOfArchiveUnitRefId"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAAltReferencesArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = "SEDAAltReferencesArchiveUnitRefId"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAFirstName(SEDAAnyEntity):
    __regid__ = "SEDAFirstName"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "first_name", "user_annotation"]
    )
    value_attr = "first_name"


class SEDABirthName(SEDAAnyEntity):
    __regid__ = "SEDABirthName"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "birth_name", "user_annotation"]
    )
    value_attr = "birth_name"


class SEDAFullName(SEDAAnyEntity):
    __regid__ = "SEDAFullName"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "full_name", "user_annotation"]
    )
    value_attr = "full_name"


class SEDAGivenName(SEDAAnyEntity):
    __regid__ = "SEDAGivenName"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "given_name", "user_annotation"]
    )
    value_attr = "given_name"


class SEDAGender(SEDAAnyEntity):
    __regid__ = "SEDAGender"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "gender", "user_annotation"]
    )
    value_attr = "gender"


class SEDABirthDate(SEDAAnyEntity):
    __regid__ = "SEDABirthDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDABirthPlace(SEDAAnyEntity):
    __regid__ = "SEDABirthPlace"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADeathDate(SEDAAnyEntity):
    __regid__ = "SEDADeathDate"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDADeathPlace(SEDAAnyEntity):
    __regid__ = "SEDADeathPlace"
    fetch_attrs, cw_fetch_order = fetch_config(["user_cardinality"])
    value_attr = None


class SEDANationality(SEDAAnyEntity):
    __regid__ = "SEDANationality"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "nationality", "user_annotation"]
    )
    value_attr = "nationality"


class SEDAExternalReference(SEDAAnyEntity):
    __regid__ = "SEDAExternalReference"
    fetch_attrs, cw_fetch_order = fetch_config(["external_reference"])
    value_attr = "external_reference"


class SEDARepositoryArchiveUnitPID(SEDAAnyEntity):
    __regid__ = "SEDARepositoryArchiveUnitPID"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDARepositoryObjectPID(SEDAAnyEntity):
    __regid__ = "SEDARepositoryObjectPID"
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None


class SEDAAddress(SEDAAnyEntity):
    __regid__ = "SEDAAddress"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "address", "user_annotation"]
    )
    value_attr = "address"


class SEDACity(SEDAAnyEntity):
    __regid__ = "SEDACity"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "city", "user_annotation"]
    )
    value_attr = "city"


class SEDACountry(SEDAAnyEntity):
    __regid__ = "SEDACountry"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "country", "user_annotation"]
    )
    value_attr = "country"


class SEDAGeogname(SEDAAnyEntity):
    __regid__ = "SEDAGeogname"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "geogname", "user_annotation"]
    )
    value_attr = "geogname"


class SEDAPostalCode(SEDAAnyEntity):
    __regid__ = "SEDAPostalCode"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "postal_code", "user_annotation"]
    )
    value_attr = "postal_code"


class SEDARegion(SEDAAnyEntity):
    __regid__ = "SEDARegion"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["user_cardinality", "region", "user_annotation"]
    )
    value_attr = "region"


CHOICE_RTYPE = {
    "SEDAAltAccessRulePreventInheritance": [
        ["seda_prevent_inheritance", "object"],
        ["seda_ref_non_rule_id_from", "object"],
    ],
    "SEDAAltAgentCorpname": [
        ["seda_corpname", "object"],
        ["seda_seq_alt_agent_corpname_first_name", "subject"],
    ],
    "SEDAAltAppraisalRulePreventInheritance": [
        ["seda_prevent_inheritance", "object"],
        ["seda_ref_non_rule_id_from", "object"],
    ],
    "SEDAAltArchiveUnitArchiveUnitRefId": [
        ["seda_archive_unit_ref_id_from", "object"],
        ["seda_seq_alt_archive_unit_archive_unit_ref_id_management", "subject"],
    ],
    "SEDAAltBinaryDataObjectAttachment": [
        ["seda_attachment", "object"],
        ["seda_uri", "object"],
    ],
    "SEDAAltClassificationRulePreventInheritance": [
        ["seda_prevent_inheritance", "object"],
        ["seda_ref_non_rule_id_from", "object"],
    ],
    "SEDAAltDisseminationRulePreventInheritance": [
        ["seda_prevent_inheritance", "object"],
        ["seda_ref_non_rule_id_from", "object"],
    ],
    "SEDAAltHoldRulePreventInheritance": [
        ["seda_prevent_inheritance", "object"],
        ["seda_ref_non_rule_id_from", "object"],
    ],
    "SEDAAltIsPartOfArchiveUnitRefId": [
        ["seda_archive_unit_ref_id_from", "object"],
        ["seda_external_reference", "object"],
        ["seda_repository_archive_unit_pid", "object"],
        ["seda_repository_object_pid", "object"],
        ["seda_data_object_reference", "object"],
    ],
    "SEDAAltIsVersionOfArchiveUnitRefId": [
        ["seda_archive_unit_ref_id_from", "object"],
        ["seda_data_object_reference", "object"],
        ["seda_external_reference", "object"],
        ["seda_repository_archive_unit_pid", "object"],
        ["seda_repository_object_pid", "object"],
    ],
    "SEDAAltReferencesArchiveUnitRefId": [
        ["seda_archive_unit_ref_id_from", "object"],
        ["seda_external_reference", "object"],
        ["seda_repository_archive_unit_pid", "object"],
        ["seda_repository_object_pid", "object"],
        ["seda_data_object_reference", "object"],
    ],
    "SEDAAltReplacesArchiveUnitRefId": [
        ["seda_archive_unit_ref_id_from", "object"],
        ["seda_external_reference", "object"],
        ["seda_repository_archive_unit_pid", "object"],
        ["seda_repository_object_pid", "object"],
        ["seda_data_object_reference", "object"],
    ],
    "SEDAAltRequiresArchiveUnitRefId": [
        ["seda_archive_unit_ref_id_from", "object"],
        ["seda_external_reference", "object"],
        ["seda_repository_archive_unit_pid", "object"],
        ["seda_repository_object_pid", "object"],
        ["seda_data_object_reference", "object"],
    ],
    "SEDAAltReuseRulePreventInheritance": [
        ["seda_prevent_inheritance", "object"],
        ["seda_ref_non_rule_id_from", "object"],
    ],
    "SEDAAltSenderCorpname": [
        ["seda_corpname", "object"],
        ["seda_seq_alt_sender_corpname_first_name", "subject"],
    ],
    "SEDAAltStorageRulePreventInheritance": [
        ["seda_prevent_inheritance", "object"],
        ["seda_ref_non_rule_id_from", "object"],
    ],
    "SEDAAltTransmitterCorpname": [
        ["seda_corpname", "object"],
        ["seda_seq_alt_transmitter_corpname_first_name", "subject"],
    ],
}
CHECK_CARD_ETYPES = [
    "SEDAActivity",
    "SEDAAddressee",
    "SEDAAgent",
    "SEDAArchivalAgencyArchiveUnitIdentifier",
    "SEDAArchiveUnit",
    "SEDAAuthorizedAgent",
    "SEDABinaryDataObject",
    "SEDACustodialHistoryItem",
    "SEDADataObjectGroup",
    "SEDADataObjectReference",
    "SEDAEvent",
    "SEDAEventAbstract",
    "SEDAFilePlanPosition",
    "SEDAFunction",
    "SEDAIdentifier",
    "SEDAIsPartOf",
    "SEDAIsVersionOf",
    "SEDAJuridictional",
    "SEDAKeyword",
    "SEDALanguage",
    "SEDALinkingAgentIdentifier",
    "SEDAMandate",
    "SEDANationality",
    "SEDAOriginatingAgencyArchiveUnitIdentifier",
    "SEDAOriginatingSystemId",
    "SEDAPhysicalDataObject",
    "SEDAPosition",
    "SEDARecipient",
    "SEDARefNonRuleId",
    "SEDAReferences",
    "SEDARelatedTransferReference",
    "SEDARelationship",
    "SEDAReplaces",
    "SEDARequires",
    "SEDARole",
    "SEDASender",
    "SEDASpatial",
    "SEDASystemId",
    "SEDATag",
    "SEDATemporal",
    "SEDATextContent",
    "SEDATransferringAgencyArchiveUnitIdentifier",
    "SEDATransmitter",
    "SEDAWriter",
]
CHECK_CHILDREN_CARD_RTYPES = [
    "seda_activity",
    "seda_addressee_from",
    "seda_agent",
    "seda_archival_agency_archive_unit_identifier",
    "seda_archive_unit",
    "seda_authorized_agent_from",
    "seda_binary_data_object",
    "seda_custodial_history_item",
    "seda_data_object_group",
    "seda_data_object_reference",
    "seda_event",
    "seda_event_abstract",
    "seda_file_plan_position",
    "seda_function",
    "seda_identifier",
    "seda_is_part_of",
    "seda_is_version_of",
    "seda_juridictional",
    "seda_keyword",
    "seda_language_from",
    "seda_linking_agent_identifier",
    "seda_mandate",
    "seda_nationality",
    "seda_originating_agency_archive_unit_identifier",
    "seda_originating_system_id",
    "seda_physical_data_object",
    "seda_position",
    "seda_recipient_from",
    "seda_ref_non_rule_id_from",
    "seda_references",
    "seda_related_transfer_reference",
    "seda_relationship",
    "seda_replaces",
    "seda_requires",
    "seda_role",
    "seda_sender",
    "seda_spatial",
    "seda_system_id",
    "seda_tag",
    "seda_temporal",
    "seda_text_content",
    "seda_transferring_agency_archive_unit_identifier",
    "seda_transmitter",
    "seda_writer_from",
]



# copyright 2016-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Extract concepts scheme as Linked CSV from XSD file.

XSD parsing is done using generateDS, which has been copied into the `gends` directory (only the
used bits).
"""
import csv
import os.path as osp

from .xsd import seda_xsd, un_camel_case

import sys

sys.exit("broken until pyxst stop using a set for textual_content_values")


def generate_csv(output_dir):
    xschema = seda_xsd()
    for xstype, attr_or_idx in (
        ("Diameter", "unit"),
        ("Weight", "unit"),
        ("FinalAction", 0),
        ("FinalAction", 1),
        ("KeywordType", 0),
        ("DescriptionLevel", 0),
    ):
        xselements = xschema.elts_index[
            "{fr:gouv:culture:archivesdefrance:seda:v2.2}" + xstype
        ]
        if isinstance(attr_or_idx, int):
            xselement = xselements[attr_or_idx]
        else:
            assert len(xselements) == 1
            xselement = xselements[0].attributes[attr_or_idx].target
        with open(osp.join(output_dir, un_camel_case(xstype) + ".csv"), "w") as stream:
            writer = csv.writer(
                stream, delimiter=";", quotechar='"', quoting=csv.QUOTE_MINIMAL
            )
            _GENERATORS[f"{xstype}_{attr_or_idx}"](
                writer, xselement.textual_content_values
            )


def _generate_pref_alt(writer, values):
    writer.writerow(["#", "$id", "label", "alt label"])
    writer.writerow(["type", "url", "string", "string"])
    writer.writerow(["lang", "", "en", "en"])
    writer.writerow(["url", "skos:Concept", "skos:prefLabel", "skos:altLabel"])

    def writerow(row):
        assert 1 <= len(row) <= 2, len(row)
        if len(row) == 1:
            row.append("")
        writer.writerow(["", ""] + row)

    row = []
    for value in values:
        if not row or value.isupper():
            row.append(value)
        else:
            writerow(row)
            row = [value]
    writerow(row)


def _generate_pref(writer, values):
    writer.writerow(["#", "$id", "label"])
    writer.writerow(["type", "url", "string"])
    writer.writerow(["lang", "", "en"])
    writer.writerow(["url", "skos:Concept", "skos:prefLabel"])
    for value in values:
        writer.writerow(["", "", value])


_GENERATORS = {
    "Diameter_unit": _generate_pref_alt,
    "Weight_unit": _generate_pref_alt,
    "FinalAction_0": _generate_pref,
    "FinalAction_1": _generate_pref,
    "KeywordType_0": _generate_pref,
    "DescriptionLevel_0": _generate_pref,
}

if __name__ == "__main__":
    import sys

    if len(sys.argv) == 2:
        output_dir = sys.argv[1]
    elif len(sys.argv) < 2:
        output_dir = "."
    else:
        raise Exception("too much arguments")
    generate_csv(output_dir)



# pylint: disable=W0622
"""cubicweb-seda application packaging information"""

modname = "seda"
distname = "cubicweb-seda"

numversion = (2, 2, 0)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
author = "LOGILAB S.A. (Paris, FRANCE)"
author_email = "contact@logilab.fr"
description = "Data Exchange Standard for Archival"
web = f"https://forge.extranet.logilab.fr/cubicweb/cubes/{distname}"

__depends__ = {
    "cubicweb": ">= 3.38.0, < 3.39.0",
    "cubicweb-eac": ">= 0.8.3, < 0.9.0",
    "cubicweb-skos": ">= 2.0.0",
    "cubicweb-compound": ">= 0.7",
    "cubicweb-relationwidget": ">= 0.4",
    "cubicweb-squareui": None,
    "cubicweb-geocoding": "< 0.4",
    "pyxst": ">= 0.3.2",
    "rdflib": ">= 4.1",
}
__recommends__ = {}

classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python",
    "Programming Language :: JavaScript",
]



from cubicweb_seda.dataimport import import_seda_schemes

rql('SET X compat_list CL + ", RNG" WHERE X compat_list CL')
commit(ask_confirm=False)

rql(
    'INSERT SEDAMimeType X: X user_cardinality "0..1", X seda_mime_type_from BDO '
    "WHERE NOT Y seda_mime_type_from BDO"
)

rql(
    'INSERT SEDAFormatId X: X user_cardinality "0..1", X seda_format_id_from BDO '
    "WHERE NOT Y seda_format_id_from BDO"
)
commit(ask_confirm=False)

sync_schema_props_perms("seda_mime_type_to")
sync_schema_props_perms("seda_format_id_to")
sync_schema_props_perms("seda_data_object_reference_id")

rql(
    "DELETE CS scheme_relation_type RT WHERE "
    'RT name IN ("seda_language_to", "seda_description_language_to")'
)
commit(ask_confirm=False)
import_seda_schemes(cnx)



from cubicweb_seda.dataimport import import_seda_schemes, LCSV_FILES

add_entity_type("SEDAMessageDigestAlgorithmCodeListVersion")
drop_relation_type("seda_message_digest_algorithm_code_list_version")
add_entity_type("SEDAFileFormatCodeListVersion")
drop_relation_type("seda_file_format_code_list_version")

add_entity_type("SEDAAcquisitionInformationCodeListVersion")
add_entity_type("SEDAAcquisitionInformation")

add_entity_type("SEDAOriginatingAgencyIdentifier")
add_entity_type("SEDASubmissionAgencyIdentifier")

add_entity_type("SEDALegalStatus")

sync_schema_props_perms("seda_algorithm")
sync_schema_props_perms("seda_format_id_to")

for etype in (
    "SEDArestrictionRuleIdRef",
    "SEDArestrictionValue",
    "SEDArestrictionEndDate",
):
    drop_entity_type(etype)

lcsv_files = [
    file_def
    for file_def in LCSV_FILES
    if file_def[-1]
    in ("legal_status.csv", "digest_algorithms.csv", "dissemination.csv")
]
import_seda_schemes(cnx, lcsv_files=lcsv_files)

drop_relation_type("seda_reply_code_list_version")
add_relation_definition("SEDAArchiveTransfer", "clone_of", "SEDAArchiveTransfer")



# copyright 2015-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda custom fields/widgets"""


from logilab.common.decorators import monkeypatch

from cubicweb import tags, utils
from cubicweb.predicates import specified_etype_implements
from cubicweb.uilib import js
from cubicweb_web import formfields as ff, formwidgets as fw
from cubicweb_web.views import ajaxcontroller, autoform, formrenderers

from cubicweb_relationwidget import views as rwdg

from cubicweb_seda import SEDA_2_2_ONLY_ETYPES

# deactivate relation widget's creation form by default, it causes some js error if e.g. there are
# some calendar widgets. Also, we don't really want that from an UX POV, IMO.
rwdg.SearchForRelatedEntitiesView.has_creation_form = False


def configure_relation_widget(req, div, search_url, title, multiple, validate):
    """Build a javascript link to invoke a relation widget

    Widget will be linked to div `div`, with a title `title`. It will display selectable entities
    matching url `search_url`. bool `multiple` indicates whether several entities can be selected or
    just one, `validate` identifies the javascript callback that must be used to validate the
    selection.
    """
    req.add_js(
        (
            "jquery.ui.js",
            "cubicweb.ajax.js",
            "cubicweb.widgets.js",
            "cubicweb.facets.js",
            "cubicweb_relationwidget.js",
            "cubes.editionext.js",
        )
    )
    return "javascript: %s" % js.editext.relateWidget(
        div, search_url, title, multiple, utils.JSString(validate)
    )


class ConceptOrTextField(ff.Field):
    """Compound field, using the :class:`ConceptAutoCompleteWidget` to handle values which may be
    either a relation to a `Concept` or a string

    It's expected to be set on the text attribute. Also, as you should prefer specifying a field's
    class, you will usually sublcass this field to set the following attributes:

    * 'scheme_relation', relation whose object will be the selected scheme
    * 'concept_relation', relation whose object will be the linked concept

    Example::

        class KeywordConceptOrTextField(ConceptOrTextField):
            scheme_relation = 'keyword_scheme'
            concept_relation = 'keyword_concept'

            def __init__(self, **kwargs):
                super(EquivalentConceptOrTextField, self).__init__(required=True, **kwargs)
                self.help = _('when linked to a vocabulary, value is enforced to the label of a '
                              'concept in this vocabulary. Remove it if you want to type text '
                              'freely.')

        aff = uicfg.autoform_field
        aff.tag_attribute(('Keyword', 'keyword_value'), KeywordConceptOrTextField)

    You may also configure the ajax function that will be used for the completion values, by
    specifying it using the 'ajax_autocomplete_func' attribute (or by overriding this ajax
    func). Default is :func:`scheme_concepts_autocomplete`.
    """

    scheme_relation = concept_relation = None
    ajax_autocomplete_func = "scheme_concepts_autocomplete"

    def get_widget(self, form):
        return ConceptAutoCompleteWidget(
            self.name, self.scheme_relation, self.ajax_autocomplete_func, optional=True
        )

    def has_been_modified(self, form):
        return True  # handled in process_posted below

    def process_posted(self, form):
        posted = form._cw.form
        text_val = posted.get(self.input_name(form, "Label"), "").strip()
        equivalent_eid = posted.get(self.input_name(form), "").strip()
        equivalent_eids = set()
        if equivalent_eid:
            equivalent_eids.add(int(equivalent_eid))
        if self.required and not (text_val or equivalent_eid):
            raise ff.ProcessFormError(form._cw.__("required field"))
        entity = form.edited_entity
        if not entity.has_eid() or getattr(entity, self.name) != text_val:
            yield (ff.Field(name=self.name, role="subject", eidparam=True), text_val)
        if (not entity.has_eid() and equivalent_eids) or (
            entity.has_eid()
            and {x.eid for x in entity.equivalent_concept} != equivalent_eids
        ):
            subfield = ff.Field(
                name=self.concept_relation, role="subject", eidparam=True
            )
            # register the association between the value and the field, because on entity creation,
            # process_posted will be recalled on the newly created field, and if we don't do that it
            # won't find the proper value (this is very nasty)
            form.formvalues[(subfield, form)] = equivalent_eids
            yield (subfield, equivalent_eids)


class KeywordTypeMasterWidget(fw.Select):
    """
    Usage::

      affk = uicfg.autoform_field_kwargs
      affk.set_field_kwargs('KeywordType', 'keyword_type_to',
                            widget=widgets.KeywordTypeMasterWidget(
                            slave_base_name='seda_keyword_reference_to_scheme'))
    """

    def __init__(self, slave_base_name, **kwargs):
        super().__init__(**kwargs)
        self.slave_base_name = slave_base_name

    def _render(self, form, field, render):
        req = form._cw

        vocabularies_data = []
        for eid, title, uri, keyword_type in req.execute(
            "Any CS,CST,CSU,CSKT WHERE CS title CST, CS cwuri CSU, CS code_keyword_type CSKT?"
        ):
            vocabularies_data.append(
                {"eid": eid, "title": title or uri, "keyword_type": keyword_type}
            )
        vocabularies_data.sort(key=lambda x: x["title"])

        req.add_js(("cubicweb.js", "cubicweb.ajax.js", "cubes.skoscomplete.js"))
        req.add_onload(
            js.typed_vocabularies.initKeywordTypeMasterWidget(
                field.dom_id(form, self.suffix), self.slave_base_name, vocabularies_data
            )
        )

        return super()._render(form, field, render)


class ConceptAutoCompleteWidget(fw.TextInput):
    """Derive from simple text input to create an autocompletion widget if a scheme is specified,
    otherwise free text is fine if `optional` argument is true. In such case:

    * `slave_name` is expected to be the name of the concept attribute, or the text attribute if
      optional is true),

    * you'll have to use this widget from a custom field that will handle the relation to the
      concept (e.g. :class:`ConceptOrTextField`).

    When optional is false, a regular :class:`RelationField` is fine.

    You may configure the ajax function that will be used for the completion values by specifying it
    using the 'ajax_autocomplete_func' argument (or by overriding this ajax func). Default is
    :func:`scheme_concepts_autocomplete`.

    Usage::

      affk = uicfg.autoform_field_kwargs
      affk.set_field_kwargs('Keyword', 'keyword_value',
                            widget=ConceptAutoCompleteWidget(slave_name='keyword_value',
                                                             master_name='keyword_scheme',
                                                             optional=True))
    """

    needs_css = ("jquery.ui.css",)
    needs_js = (
        "jquery.ui.js",
        "cubicweb.js",
        "cubicweb.ajax.js",
        "cubes.skoscomplete.js",
    )

    def __init__(
        self,
        slave_name,
        master_name,
        ajax_autocomplete_func="scheme_concepts_autocomplete",
        optional=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.slave_name = slave_name
        self.master_name = master_name
        self.ajax_autocomplete_func = ajax_autocomplete_func
        self.optional = optional

    def _render(self, form, field, render):
        entity = form.edited_entity
        slave_id = field.dom_id(form, self.suffix)
        master_id = slave_id.replace(self.slave_name, self.master_name)
        if entity.has_eid():
            concept = entity.concept
        else:
            concept = None
        req = form._cw
        req.add_onload(
            js.concept_autocomplete.initConceptAutoCompleteWidget(
                master_id, slave_id, self.ajax_autocomplete_func
            )
        )
        if concept is None:
            value = getattr(entity, self.slave_name) if self.optional else None
            eid = ""
        else:
            value = concept.label()
            eid = str(concept.eid)
        # we need an hidden input to handle the value while the text input display the label
        inputs = [
            tags.input(
                name=field.input_name(form, "Label"),
                id=slave_id + "Label",
                klass="form-control",
                type="text",
                value=value,
            ),
            tags.input(
                name=field.input_name(form), id=slave_id, type="hidden", value=eid
            ),
        ]
        return "\n".join(inputs)


@ajaxcontroller.ajaxfunc(output_type="json")
def scheme_concepts_autocomplete(self):
    assert self._cw.form["scheme"]
    scheme = int(self._cw.form["scheme"])
    term = self._cw.form["q"]
    limit = self._cw.form.get("limit", 50)
    return [
        {"value": eid, "label": label}
        for eid, label in self._cw.execute(
            "DISTINCT Any C,N ORDERBY N LIMIT %s WHERE C in_scheme S, S eid %%(s)s, "
            "C preferred_label L, L label N, L label ILIKE %%(term)s" % limit,
            {"s": scheme, "term": f"%{term}%"},
        )
    ]


@monkeypatch(formrenderers.EntityInlinedFormRenderer)
def render_title(self, w, form, values):
    """Monkey-patched to remove counter"""
    w('<div class="iformTitle">')
    w(f"<span>{values['title']}</span>")
    if form.edited_entity.cw_etype in SEDA_2_2_ONLY_ETYPES:
        w("&nbsp;<span class='badge badge-warning'>SEDA 2.2</span>")
    if values["removejs"]:
        values["removemsg"] = self._cw.__("remove-inlined-entity-form")
        w(
            f" [<a href=\"javascript: {values['removejs']};$.noop();\">{values['removemsg']}</a>]"
        )
    w("</div>")


class SEDAMetaField(ff.StringField):
    """Extends user_cardinality field to allow edition of a `user_annotation` all together."""

    @staticmethod
    def annotation_field(form):
        """Return the field for `user_annotation` entity."""
        eschema = form.edited_entity.e_schema
        rschema = eschema.subjrels["user_annotation"]
        return ff.guess_field(eschema, rschema, "subject", form._cw)

    def actual_fields(self, form):
        """Overriden from :class:`Field`"""
        yield self
        yield self.annotation_field(form)

    def render(self, form, renderer):
        """Overriden from :class:`Field` to render fields in a div which is hidden by default"""
        form._cw.add_js("cubes.seda.js")
        wdgs = [self.get_widget(form).render(form, self, renderer)]
        self._render_hidden_section(wdgs, form, renderer)
        return "\n".join(wdgs)

    def _render_hidden_section(self, wdgs, form, renderer):
        divid = f"{self.input_name(form)}-advanced"
        if form.edited_entity.has_eid() and form.edited_entity.user_annotation:
            hidden = ""
            icon = "icon-up-open"
        else:
            hidden = " hidden"
            icon = "icon-list-add"
        wdgs.append(
            tags.a(
                "",
                onclick=str(js.seda.toggleFormMetaVisibility(divid)),
                href="javascript:$.noop()",
                title=form._cw._("show/hide meta fields"),
                # take care, related js relies on the icon class position
                klass=icon + " metaFieldSwitch",
            )
        )
        wdgs.append(f'<div id="{divid}" class="metaField{hidden}">')
        wdgs.append(self._render_subfield(form, self.annotation_field(form), renderer))
        wdgs.append("</div>")

    @staticmethod
    def _render_subfield(form, field, renderer):
        """Render a sub-field: label + widget + help + EOL"""
        data = utils.UStringIO()
        w = data.write
        w('<div class="row">')
        w(renderer.render_label(form, field))
        w('<div class="col-md-9">')
        w(field.render(form, renderer))
        w("</div>")
        w("</div>")
        w('<div class="row">')
        w('<div class="col-md-offset-3 col-md-9">')
        w(renderer.render_help(form, field))
        w("</div>")
        w("</div>")
        return data.getvalue()


class SimplifiedAutomaticEntityForm(autoform.AutomaticEntityForm):
    """Custom autoform, forcing display of creation form instead of add new link."""

    __abstract__ = True

    def should_display_inline_creation_form(self, rschema, existing, card):
        return not existing

    def should_display_add_new_relation_link(self, rschema, existing, card):
        return False


class NoTitleEntityInlinedFormRenderer(formrenderers.EntityInlinedFormRenderer):
    """Custom inlined form renderer that doesn't display any title nor remove link.

    This is intended to be subclassed with a custom selector.
    """

    __abstract__ = True

    def render_title(self, w, form, values):
        pass


class ConcretNoTitleEntityInlinedFormRenderer(NoTitleEntityInlinedFormRenderer):
    """Concret implementation of `NoTitleEntityInlinedFormRenderer` with a custom regid.

    Use this one by specifying the renderer id explicitly, for case where you can't easily specify a
    selector.
    """

    __regid__ = "notitle"


class SEDAWarningExportInlineAddNewLinkView(autoform.InlineAddNewLinkView):
    """
    Add export warning tag on SEAD 2.2 relations
    """

    __select__ = autoform.InlineAddNewLinkView.__select__ & specified_etype_implements(
        *SEDA_2_2_ONLY_ETYPES
    )

    def call(self, i18nctx, **kwargs):
        self._cw.set_varmaker()
        divid = "addNew%s%s%s:%s" % (self.etype, self.rtype, self.role, self.peid)
        self.w('<div class="inlinedform" id="%s" cubicweb:limit="true">', divid)
        js = "addInlineCreationForm('%s', '%s', '%s', '%s', '%s', '%s')" % (
            self.peid,
            self.petype,
            self.etype,
            self.rtype,
            self.role,
            i18nctx,
        )
        if self.pform.should_hide_add_new_relation_link(self.rtype, self.card):
            js = "toggleVisibility('%s'); %s" % (divid, js)
        __ = self._cw.pgettext
        self.w(
            '<a class="addEntity" id="add%s:%slink" href="javascript: %s" >+ %s.</a>',
            self.rtype,
            self.peid,
            js,
            __(i18nctx, "add a %s" % self.etype),
        )
        self.w("<span class='badge badge-warning inlined'>SEDA 2.2</span>")
        self.w("</div>")



sync_schema_props_perms("AuthorityRecord")

add_relation_type("wrap_dataobjects")

# XXX restore relations to [Binary|Physical]DataObjects

for group in rql("Any X WHERE X is SEDADataObjectGroup").entities():
    archive = group.seda_data_object_group
    if len(archive) != 1:
        continue
    archive = archive[0]
    if archive.cw_etype == "SEDAArchiveTransfer":
        archive.cw_set(
            reverse_seda_binary_data_object=group.reverse_seda_binary_data_object,
            reverse_seda_physical_data_object=group.reverse_seda_physical_data_object,
        )
        group.cw_delete()
        cnx.commit()


print(
    f"found {rql('Any COUNT(X) WHERE X is SEDADataObjectGroup')[0][0]} SEDADataObjectGroup(s) left"
)



scheme = cnx.find("ConceptScheme", title="Types MIME").one()
sql(
    f"""
UPDATE cw_label SET cw_language_code = 'seda'
FROM cw_concept, in_scheme_relation
WHERE in_scheme_relation.eid_to = {scheme.eid}
AND cw_concept.cw_eid = in_scheme_relation.eid_from
AND cw_label_of = cw_concept.cw_eid
AND cw_language_code = 'seda-2'
"""
)
